/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */



#include <sys/ioctl.h>
#include <net/if.h>
#include <unistd.h>
#include <errno.h>

#include "private.h"

typedef struct iface_s {
	struct sockaddr_in ip_addr;
	struct sockaddr_in bcast_addr;
	int8_t hw_addr[6] ;
	struct iface_s *next;
} iface_t ;


#define INITIAL_IFACE_COUNT 10
#define IFACE_COUNT_INC 5

// returns a linked list of interfaces
iface_t *get_ifaces(int sd) {
	struct ifconf ifc;
	struct ifreq *ifr, ifrcopy ;
	struct sockaddr_in *sin ;
	int len, lastlen, flags ;
	char *buf, *ptr ;
	iface_t *ift_head, **ift_next, *ift ;

	ift_head = NULL ;
	ift_next = &ift_head ;

	// first use ioctl to get a listing of interfaces 
	lastlen = 0; 
	len = INITIAL_IFACE_COUNT * sizeof(struct ifreq) ;

	for (;;) {
		buf = Malloc(len) ;
		ifc.ifc_len = len ;
		ifc.ifc_buf = buf ;
		if ( ioctl(sd, SIOCGIFCONF, &ifc) < 0 ) {
			if ( errno != EINVAL || lastlen != 0 ) {
				err_fatal("ioctl error") ;
			}
		} else {
			if(ifc.ifc_len == lastlen) 
				break ;
			lastlen = ifc.ifc_len ;
		}
		len += IFACE_COUNT_INC ;
		free(buf) ;
	}

	// loop through each iface
	for(ptr = buf ; ptr < buf + ifc.ifc_len ; ) {
		ifr = (struct ifreq* ) ptr ;
		
		// work out length here
#ifdef	HAVE_SOCKADDR_SA_LEN
		len = max(sizeof( SA), ifr->ifr_addr.sa_len) ;
#else
		switch (ifr->ifr_addr.sa_family) {
#ifdef	IPV6
			case AF_INET6:
				len = sizeof(struct sockaddr_in6) ;
				break;
#endif
			case AF_INET:
			default:
				len = sizeof(SA) ;
				break ;
		}
#endif /* HAVE_SOCKADDR_SA_LEN */

		ptr += sizeof(ifr->ifr_name) + len ;

		// look for AF_INET interfaces
		if(ifr->ifr_addr.sa_family == AF_INET) {
			ifrcopy = *ifr ;
			if (ioctl(sd, SIOCGIFFLAGS, &ifrcopy) < 0) 
				err_fatal("ioctl failed") ;
		
			flags = ifrcopy.ifr_flags ;
			if( (flags & IFF_UP) == 0) 
				continue ; //skip down interfaces

			if( (flags & IFF_LOOPBACK) ) 
				continue ; //skip lookback
			
			// interesting iface, better malloc for it ..
			ift = Calloc(1, sizeof(iface_t) );

			if(ift_head == NULL) {
				ift_head = ift ;
			} else {
				*ift_next = ift ;
			}
			
			ift_next = &ift->next ;

			sin = (struct sockaddr_in *) &ifr->ifr_addr ;
			ift->ip_addr.sin_addr = sin->sin_addr ;

			// fetch bcast address
#ifdef SIOCGIFBRDADDR
			if(flags & IFF_BROADCAST) {
				if (ioctl(sd, SIOCGIFBRDADDR, &ifrcopy) < 0) 
					err_fatal("ioctl failed") ;
		
				sin = (struct sockaddr_in *) &ifrcopy.ifr_broadaddr ;
				ift->bcast_addr.sin_addr = sin->sin_addr ;
			}
#endif
			// fetch hardware address
#ifdef SIOCGIFHWADDR
			if(flags & SIOCGIFHWADDR) {
				if (ioctl(sd, SIOCGIFHWADDR, &ifrcopy) < 0) 
					err_fatal("ioctl failed") ;
		
				memcpy(&ift->hw_addr , ifrcopy.ifr_hwaddr.sa_data, 6) ;
			}
#endif
			
		/* ok, if that all failed we should prob try and use sysctl to work out the bcast
		 * and hware addresses
		 * i'll leave that for another day
		 */
				
		}	
	}
	free(buf) ;
	
	return ift_head ;

}

// free memory used by the iface's list
void free_ifaces(iface_t *head) {
	iface_t *ift , *ift_next  ;

	for(ift = head ; ift != NULL; ift = ift_next) {
		ift_next = ift->next ;
		free(ift) ;
	}

}


int net_init(node n, const char *ip) {
	iface_t *ift_head, *ift ;
	struct in_addr wanted_ip ;
	
	int found = FALSE ;
	int i, sd ;

	// create socket to get iface config
	sd = socket(PF_INET  ,SOCK_DGRAM , 0 ) ;
	
	if(sd <0 )
		err_fatal("Cannot open socket");

	ift_head = get_ifaces(sd) ;

	if(n->state.verbose) {
		printf("#### INTERFACES FOUND ####\n") ;
		for (ift = ift_head ;ift != NULL ; ift = ift->next ) {
			printf("IP: %s\n", inet_ntoa(ift->ip_addr.sin_addr) ) ;
			printf("  bcast: %s\n" , inet_ntoa(ift->bcast_addr.sin_addr) ) ;
			printf("  hwaddr: ") ;
				for(i = 0; i < 6; i++ ) {
					printf("%hhx:", ift->hw_addr[i] ) ;
				}
			printf("\n") ;
		}
		printf("#########################\n") ;
	}

	if( ip == NULL ) {
		// pick first address
		// copy ip address, bcast address and hardware address
		n->state.ip_addr = ift_head->ip_addr.sin_addr ;
		n->state.bcast_addr = ift_head->bcast_addr.sin_addr ;
		memcpy(&n->state.hw_addr, &ift_head->hw_addr, 6 );

	} else {
		// search through list of interfaces for one with the correct address
		if(inet_aton(ip, &wanted_ip) < 0 ) 
			err_fatal("Cannot convert address %s\n", ip) ;

		for (ift = ift_head; ift != NULL; ift = ift->next ) {
			if( ift->ip_addr.sin_addr.s_addr == wanted_ip.s_addr) {
				found = TRUE ;
				n->state.ip_addr = ift->ip_addr.sin_addr ;
				n->state.bcast_addr = ift->bcast_addr.sin_addr ;
				memcpy(&n->state.hw_addr, &ift->hw_addr, 6 );

			}
		}
		if(! found) 
			err_fatal("Cannot find ip %s\n", ip) ;
	}
	close(sd) ;
	free_ifaces(ift_head) ;

	return 0 ;

}

int net_start(node n) {
	struct sockaddr_in servAddr ;
	int bcast_flag = TRUE ;
	node tmp ;
	
	/* create socket */
	n->sd[0] = socket(PF_INET  ,SOCK_DGRAM , 0 ) ;
	
	if(n->sd[0] <0)
		err_fatal("Cannot open sockets");

	servAddr.sin_family = AF_INET ;
	servAddr.sin_port = htons(ARTNET_PORT) ;
	
	servAddr.sin_addr.s_addr = n->state.ip_addr.s_addr ;

	if( n->state.verbose) 
		printf("Binding to %s \n" , inet_ntoa(servAddr.sin_addr)  );
	
	/* bind sockets 
	 * we do one for the ip address and one for the broadcast address
	 */
	if( bind(n->sd[0], (SA *) &servAddr, sizeof(servAddr)) < 0  ) {
		err_fatal("falied to bind to socket") ;
	}

	// only attempt to bind to the broadcast if we are the group master
	if(n->peering.master == TRUE) {
		n->sd[1] = socket(PF_INET  ,SOCK_DGRAM , 0 ) ;

		if(n->sd[1] < 0)
			err_fatal("Cannot open sockets");

		servAddr.sin_addr.s_addr = n->state.bcast_addr.s_addr ;

		if(n->state.verbose) 
			printf("binding to %s \n" , inet_ntoa(servAddr.sin_addr)  );

		if( bind(n->sd[1], (SA *) &servAddr, sizeof(servAddr)) < 0  ) {
			err_fatal("falied to bind to socket") ;
		}
		
		// now we need to propagate sd[1] to all our peers
		for( tmp = n->peering.peer ; tmp != NULL && tmp != n ; tmp = tmp->peering.peer ) 
			tmp->sd[1] = n->sd[1] ;

	}

	// allow bcasting
	setsockopt(n->sd[0], SOL_SOCKET, SO_BROADCAST, &bcast_flag, sizeof(int) ) ;

	return 0 ;
}

int net_set_fdset(node n, fd_set *fdset) {
 	int maxfdp1 ;
 	
 	FD_SET(n->sd[0], fdset) ;
 	FD_SET(n->sd[1], fdset) ;
 	
 	maxfdp1 = max(n->sd[0] , n->sd[1]) + 1;
 	return maxfdp1 ;
}


ssize_t net_recv(node n, artnet_packet  p, int delay) {
	ssize_t     len;
	struct sockaddr_in cliAddr ;
	int cliLen = sizeof(cliAddr) ;
	fd_set rset ;
	struct timeval tv ;
	int active_sd = 0 ;
	int maxfdp1 = max(n->sd[0] , n->sd[1]) + 1;

	FD_ZERO(&rset) ;
 	maxfdp1 = net_set_fdset(n,&rset) ;

	tv.tv_usec = 0 ;
	tv.tv_sec = delay ;
	
	switch ( select( maxfdp1, &rset, NULL, NULL, &tv ) ) {
		case 0:
			// timeout
			return 0;
			break ;
		case -1:
			err_warn("select error") ;
			return 0;
			break ;
		default:
			if(FD_ISSET(n->sd[0], &rset)  )
				active_sd = n->sd[0] ;
				p->to = n->state.ip_addr;

			if(FD_ISSET(n->sd[1], &rset) )
				active_sd = n->sd[1] ;
				p->to = n->state.bcast_addr;
			
			break;
	}
	
	// need a check here for the amount of data read
	// should prob allow an extra byte after data, and pass the size as sizeof(Data) +1
	// then check the size read and if equal to size(data)+1 we have an error
	if ( (len = recvfrom(active_sd, &(p->data), sizeof(p->data), 0, (SA*) &cliAddr, &cliLen) )< 0)
        err_fatal("recvfrom error");

	if (cliAddr.sin_addr.s_addr == n->state.ip_addr.s_addr) {
		return 0;
	}

	memcpy(&(p->from), &cliAddr.sin_addr, sizeof(struct in_addr) ) ;
	// should set to in here
	
    return len;

}

int net_send(node n, artnet_packet p) {
	struct sockaddr_in addr ;
	int ret ;
	
	if (n->state.mode != ARTNET_ON) 
		return 0;

	addr.sin_family = AF_INET ;
	addr.sin_port = htons(ARTNET_PORT) ;
	addr.sin_addr = p->to ;
	p->from = n->state.ip_addr ;
	
	if(n->state.verbose) 
		printf("sending to %s\n" , inet_ntoa(addr.sin_addr) ) ;

	ret = sendto(n->sd[0], &p->data, p->length, 0, (SA*) &addr, sizeof(addr)) ;
	if ( ret == -1 ) {
		err_warn("sendto failed") ;
		n->state.report_code = ARTNET_RCUDPFAIL ;
		return 1;
	} else if (p->length != ret ) {
		err_warn("sendto failed") ;
		n->state.report_code = ARTNET_RCSOCKETWR1 ;
		return 1;
	}

	if(n->callbacks.send.fh != NULL) {
		get_type(p) ;
		n->callbacks.send.fh(n, p, n->callbacks.send.data) ;
	}
	return 0;
}


void net_reprogram(node n) {
	iface_t *ift_head, *ift ;
	int i ;

	ift_head = get_ifaces(n->sd[0]) ;

	for (ift = ift_head ;ift != NULL ; ift = ift->next ) {
		printf("IP: %s\n", inet_ntoa(ift->ip_addr.sin_addr) ) ;
		printf("  bcast: %s\n" , inet_ntoa(ift->bcast_addr.sin_addr) ) ;
		printf("  hwaddr: ") ;
			for(i = 0; i < 6; i++ ) {
				printf("%hhx:", ift->hw_addr[i] ) ;
			}
		printf("\n") ;
	}

	free_ifaces(ift_head) ;


}


int net_close(node n) {
	close(n->sd[0]) ;
	close(n->sd[1]) ;
	return 0;
}

// n2 is joining n1's group
// n1 is the group master
int net_join(node n1, node n2) {
	close(n2->sd[1]) ;
	n2->sd[1] = n1->sd[1] ;

	return 0 ;
}
